//
//------------------------------------------------------------------------------
// Name    : tone.bas
// Author  : Jerry Messina
// Date    : 7/27/2021
// Version : 1.00
// Notes   : A software digital square wave tone generation library
//------------------------------------------------------------------------------
//
// swordfish port of arduino Tone library
// based on code from https://github.com/bhagman/Tone
//  - uses TMR1, CCP1, and interrupt to generate a tone on a user-defined 
//    port pin (set by '#option TONE_OUTPUT_PIN')
//    contains 18FxxK22 device specific settings for TMR1 and CCP1
// - add blocking/non-blocking modes (see setwait and #option TONE_WAIT) 
// - add Wait() for non-blocking mode
// - add support for generating a pause by calling Play() with freq=0
// - add overloaded Play() methods for arrays of notes (freq, duration word pairs) 
// - changes 'duration' to a word parameter (65535ms) instead of 32-bit 
//
// the upper and lower freq limits depend on the system clock freq
//   clock=64MHz: lower limit = 16Hz, upper limit = 64KHz
//   clock= 4MHz: lower limit =  1Hz, upper limit = 20KHz
//
// examples:
//  tone.play(440)               ' play A (440Hz) until stop() called
//
//  tone.play(NOTE_CS4, 1000)    ' play C# (277Hz) for 1000ms
//  tone.wait()                  ' wait for tone to finish (wait_flag = false)
//
//  tone.setwait()               ' set wait_flag true... no longer need wait()
//  tone.play(277, 150)          ' play C# (277Hz) for 150ms... returns when done
//
//  const song() as word =       ' an array of (freq, duration) note pairs
//  (
//      NOTE_A4, 500,  
//      NOTE_PAUSE, 250,  
//      NOTE_F4, 350,
//      NOTE_CS4, 850
//  }    
//  tone.play(@song)
//
module tone

// specify default output pin
#option TONE_OUTPUT_PIN = PORTC.2
#option _TONE_OUTPUT_TRIS = GetTris(TONE_OUTPUT_PIN)    // for other modules (if needed) 

// default wait_flag setting
// this controls blocking/non-blocking mode operation 
// (see wait_flag, SetWait, Wait, and isPlaying)
#option TONE_WAIT = false
const TONE_WAIT_DEFAULT = TONE_WAIT

// set interrupt priority level...
#option TONE_PRIORITY = ipHigh     
#if not(TONE_PRIORITY in (ipLow, ipHigh))
   #error TONE_PRIORITY, "Invalid option... priority must be ipHigh or ipLow"
#endif
#if (TONE_PRIORITY = ipHigh)
  const TONE_PRIORITY_LEVEL = 1
#else
  const TONE_PRIORITY_LEVEL = 0
#endif

// for optional icd debugger _trap() support
#option _TONE_TRAP = false
#if (_TONE_TRAP)
  #if not defined(MPLAB_ICD) or (MPLAB_ICD = false)
    #error "_TONE_TRAP requires '#option MPLAB_ICD = true'"
  #endif
  include "mplabicd.bas"
#endif

// 88-key piano note frequencies, in Hz
//  C, C#, D, D#, E, F, F#, G, G#, A, A#, B
public const 
    NOTE_B0  as word = 31,
    NOTE_C1  as word = 33,
    NOTE_CS1 as word = 35,
    NOTE_D1  as word = 37,
    NOTE_DS1 as word = 39,
    NOTE_E1  as word = 41,
    NOTE_F1  as word = 44,
    NOTE_FS1 as word = 46,
    NOTE_G1  as word = 49,
    NOTE_GS1 as word = 52,
    NOTE_A1  as word = 55,
    NOTE_AS1 as word = 58,
    NOTE_B1  as word = 62,
    NOTE_C2  as word = 65,
    NOTE_CS2 as word = 69,
    NOTE_D2  as word = 73,
    NOTE_DS2 as word = 78,
    NOTE_E2  as word = 82,
    NOTE_F2  as word = 87,
    NOTE_FS2 as word = 93,
    NOTE_G2  as word = 98,
    NOTE_GS2 as word = 104,
    NOTE_A2  as word = 110,
    NOTE_AS2 as word = 117,
    NOTE_B2  as word = 123,
    NOTE_C3  as word = 131,
    NOTE_CS3 as word = 139,
    NOTE_D3  as word = 147,
    NOTE_DS3 as word = 156,
    NOTE_E3  as word = 165,
    NOTE_F3  as word = 175,
    NOTE_FS3 as word = 185,
    NOTE_G3  as word = 196,
    NOTE_GS3 as word = 208,
    NOTE_A3  as word = 220,
    NOTE_AS3 as word = 233,
    NOTE_B3  as word = 247,
    NOTE_C4  as word = 262,     // RTTTL (RingTone Text Transfer Language) notes start here
    NOTE_CS4 as word = 277,
    NOTE_D4  as word = 294,
    NOTE_DS4 as word = 311,
    NOTE_E4  as word = 330,
    NOTE_F4  as word = 349,
    NOTE_FS4 as word = 370,
    NOTE_G4  as word = 392,
    NOTE_GS4 as word = 415,
    NOTE_A4  as word = 440,
    NOTE_AS4 as word = 466,
    NOTE_B4  as word = 494,
    NOTE_C5  as word = 523,
    NOTE_CS5 as word = 554,
    NOTE_D5  as word = 587,
    NOTE_DS5 as word = 622,
    NOTE_E5  as word = 659,
    NOTE_F5  as word = 698,
    NOTE_FS5 as word = 740,
    NOTE_G5  as word = 784,
    NOTE_GS5 as word = 831,
    NOTE_A5  as word = 880,
    NOTE_AS5 as word = 932,
    NOTE_B5  as word = 988,
    NOTE_C6  as word = 1047,
    NOTE_CS6 as word = 1109,
    NOTE_D6  as word = 1175,
    NOTE_DS6 as word = 1245,
    NOTE_E6  as word = 1319,
    NOTE_F6  as word = 1397,
    NOTE_FS6 as word = 1480,
    NOTE_G6  as word = 1568,
    NOTE_GS6 as word = 1661,
    NOTE_A6  as word = 1760,
    NOTE_AS6 as word = 1865,
    NOTE_B6  as word = 1976,
    NOTE_C7  as word = 2093,
    NOTE_CS7 as word = 2217,
    NOTE_D7  as word = 2349,
    NOTE_DS7 as word = 2489,
    NOTE_E7  as word = 2637,
    NOTE_F7  as word = 2794,
    NOTE_FS7 as word = 2960,
    NOTE_G7  as word = 3136,
    NOTE_GS7 as word = 3322,
    NOTE_A7  as word = 3520,
    NOTE_AS7 as word = 3729,
    NOTE_B7  as word = 3951,     // last RTTTL note
    NOTE_C8  as word = 4186,
    NOTE_CS8 as word = 4435,
    NOTE_D8  as word = 4699,
    NOTE_DS8 as word = 4978,
    NOTE_PAUSE as word = 0     // special case... generates a delayms call

// enharmonic notes (black keys, denoted using '-')
//  C# = D-, D# = E-, F# = G-, G# = A-, A# = B-
public const 
    NOTE_DF1 = NOTE_CS1,    // D flat = C charp
    NOTE_EF1 = NOTE_DS1,
    NOTE_GF1 = NOTE_FS1,
    NOTE_AF1 = NOTE_GS1,
    NOTE_BF1 = NOTE_AS1,
    NOTE_DF2 = NOTE_CS2,
    NOTE_EF2 = NOTE_DS2,
    NOTE_GF2 = NOTE_FS2,
    NOTE_AF2 = NOTE_GS2,
    NOTE_BF2 = NOTE_AS2,
    NOTE_DF3 = NOTE_CS3,
    NOTE_EF3 = NOTE_DS3,
    NOTE_GF3 = NOTE_FS3,
    NOTE_AF3 = NOTE_GS3,
    NOTE_BF3 = NOTE_AS3,
    NOTE_DF4 = NOTE_CS4,
    NOTE_EF4 = NOTE_DS4,
    NOTE_GF4 = NOTE_FS4,
    NOTE_AF4 = NOTE_GS4,
    NOTE_BF4 = NOTE_AS4,
    NOTE_DF5 = NOTE_CS5,
    NOTE_EF5 = NOTE_DS5,
    NOTE_GF5 = NOTE_FS5,
    NOTE_AF5 = NOTE_GS5,
    NOTE_BF5 = NOTE_AS5,
    NOTE_DF6 = NOTE_CS6,
    NOTE_EF6 = NOTE_DS6,
    NOTE_GF6 = NOTE_FS6,
    NOTE_AF6 = NOTE_GS6,
    NOTE_BF6 = NOTE_AS6,
    NOTE_DF7 = NOTE_CS7,
    NOTE_EF7 = NOTE_DS7,
    NOTE_GF7 = NOTE_FS7,
    NOTE_AF7 = NOTE_GS7,
    NOTE_BF7 = NOTE_AS7,
    NOTE_DF8 = NOTE_CS8,
    NOTE_EF8 = NOTE_DS8
 
// piano notes array
public const notes() as word = 
(
    NOTE_B0,
    NOTE_C1, NOTE_CS1, NOTE_D1, NOTE_DS1, NOTE_E1, NOTE_F1, NOTE_FS1, NOTE_G1, NOTE_GS1, NOTE_A1, NOTE_AS1, NOTE_B1,
    NOTE_C2, NOTE_CS2, NOTE_D2, NOTE_DS2, NOTE_E2, NOTE_F2, NOTE_FS2, NOTE_G2, NOTE_GS2, NOTE_A2, NOTE_AS2, NOTE_B2,
    NOTE_C3, NOTE_CS3, NOTE_D3, NOTE_DS3, NOTE_E3, NOTE_F3, NOTE_FS3, NOTE_G3, NOTE_GS3, NOTE_A3, NOTE_AS3, NOTE_B3,
    NOTE_C4, NOTE_CS4, NOTE_D4, NOTE_DS4, NOTE_E4, NOTE_F4, NOTE_FS4, NOTE_G4, NOTE_GS4, NOTE_A4, NOTE_AS4, NOTE_B4,
    NOTE_C5, NOTE_CS5, NOTE_D5, NOTE_DS5, NOTE_E5, NOTE_F5, NOTE_FS5, NOTE_G5, NOTE_GS5, NOTE_A5, NOTE_AS5, NOTE_B5,
    NOTE_C6, NOTE_CS6, NOTE_D6, NOTE_DS6, NOTE_E6, NOTE_F6, NOTE_FS6, NOTE_G6, NOTE_GS6, NOTE_A6, NOTE_AS6, NOTE_B6,
    NOTE_C7, NOTE_CS7, NOTE_D7, NOTE_DS7, NOTE_E7, NOTE_F7, NOTE_FS7, NOTE_G7, NOTE_GS7, NOTE_A7, NOTE_AS7, NOTE_B7,
    NOTE_C8, NOTE_CS8, NOTE_D8, NOTE_DS8
)

// instruction cycle time FOSC/4 (required for using CCP)
const FCYC as longword = (_clock * 1000000) / 4

// sound output pin
dim OUTPUT_PIN as TONE_OUTPUT_PIN.TONE_OUTPUT_PIN@

// duration counter
//  > 0 - duration specified
//  = 0 - stopped
//  < 0 - infinitely (until stop() method called, or new play() called)
dim tmr_toggle_count as longint

// wait_flag (blocking/non-blocking mode)
// if true, play() is blocking and waits for tone to finish before returning
// if false, play() is non-blocking and returns immediately
// the user must call isPlaying() or wait() between tones
dim wait_flag as boolean

//-----------------------------------------------
// 18FxxK22 device specific settings
//-----------------------------------------------
// uses TMR1 and CCP1

// TMRx setup (K22 specific)
inline sub TxCON_setup()
    T1CON = (%00 << 6) +        // TMR1CS[1:0] - TMR1 clock source is instruction clock (FOSC/4)
            (%00 << 4) +        // T1CKPS[1:0] - prescaler value 1:1
            (%0 << 3) +         // T1SOSCEN - sec osc disabled
            (%0 << 2) +         // T1SYNC - ignored
            (%0 << 1) +         // T1RD16 - 8-bit rd/wr
            (%0 << 0)           // TMR1ON - TMR OFF
    TMR1H = 0
    TMR1L = 0
end sub

// TxCON TMR prescaler... either 1:1 or 1:8
macro TMR_PRESCALER(tckps)
    T1CON = (T1CON and %11001111) or (tckps << 4)   // T1CKPS[1:0] 
end macro

// start TMR
inline sub TMRx_ON()
    T1CON.bits(0) = 1           // TMR1ON
end sub

// stop TMR
inline sub TMRx_OFF()
    T1CON.bits(0) = 0           // TMR1ON
end sub

// set TMR prescaler... either 1:1 or 1:8
inline sub SET_TMRx_PRESCALER(prescaler as byte)
    TMRx_OFF()                      // stop TMR
    TMR1H = 0
    TMR1L = 0
    TMR_PRESCALER(prescaler)        // T1CKPS[1:0] 
end sub

// CCP1 interrupt bits (K22 specific)
dim CCP1IF as PIR1.bits(2),
    CCP1IE as PIE1.bits(2),
    CCP1IP as IPR1.bits(2)

// CCP1 timer select
// assign TMR1 to CCP1 (K22 specific)
inline sub assign_TMR_to_CCP()
    CCPTMRS0.bits(1) = 0        // C1TSEL[1:0] - 00=CCP1 compare mode use TMR1
    CCPTMRS0.bits(0) = 0
end sub

// CCP1 setup (K22 specific)
inline sub CCPxCON_setup()
    CCP1CON = (%0000 << 4) +    // upper bits unused
              (%1011 << 0)      // CCP1M[3:0]: compare w/special event trigger mode (CCPxIF is set, TMRx is reset)
    assign_TMR_to_CCP()         // assign TMRx to CCPx
end sub

// load CCP1 compare value
macro LOAD_CCPx_COMPARE(w)
    CCPR1H = w.bytes(1)
    CCPR1L = w.bytes(0)
    CCP1IF = 0
    CCP1IE = 1
end macro

// CCPx interrupt bit alias
dim CCPxIF as CCP1IF,
    CCPxIE as CCP1IE,
    CCPxIP as CCP1IP

//-----------------------------------------------
// end of 18FxxK22 device specific settings
//-----------------------------------------------

private inline sub clrWDT()
    asm-
        CLRWDT
    end asm
end sub
            
//-----------------------------------------------
// CCPx interrupt
//-----------------------------------------------
interrupt tone_interrupt(TONE_PRIORITY)
    if (tmr_toggle_count <> 0) then
        toggle(OUTPUT_PIN)      // toggle pin
        if (tmr_toggle_count > 0) then
            tmr_toggle_count = tmr_toggle_count - 1
        endif
    else
        CCPxIE = 0              // disable the interrupt
        OUTPUT_PIN = 0          // keep pin low after stop
    endif
    CCPxIF = 0
end interrupt

//-----------------------------------------------
// TMRx and CCPx setup
//-----------------------------------------------
// initialize TMRx and CCPx
public sub init(enable_intr as boolean = true)
    low(OUTPUT_PIN)

    // TMR setup
    TxCON_setup()

    // CCP setup
    CCPxIF = 0
    CCPxIE = 0
    CCPxCON_setup()

    // set intr priority
    CCPxIP = TONE_PRIORITY_LEVEL
    
    if (enable_intr) then
        enable(tone_interrupt)
    endif
end sub

//-----------------------------------------------
// Tone library routines
//-----------------------------------------------

// stop playing
public sub stop()
    TMRx_OFF()              // stop TMR 
    CCPxIE = 0              // disable CCP intr
    OUTPUT_PIN = 0
  #if (_TONE_TRAP)
    _trap()                 // halt debugger
  #endif
end sub

// check to see if a tone is currently playing
//  returns true = playing, false = not playing/done
public function isPlaying() as boolean
    result = boolean(CCPxIE)
end function

// set module wait_flag
public inline sub setwait(flag as boolean = true)
    wait_flag = flag
end sub

// wait until tone is done playing
// returns immediately if toggle_count is <= 0 (done or continuous)
public sub wait()
    if (tmr_toggle_count > 0) then
        while (isPlaying())
            clrWDT()
        end while
    endif
end sub

//
// play a tone (with optional duration)
// parameters: 
//   freq (in hz), 
//   duration (in msecs), 0-65535
// if duration is not specified (or is 0) then the tone plays until user calls stop()
// if duration > 0 and module 'wait_flag' is true then play() will wait for the tone 
//   to finish before returning (blocking operation), otherwise it's non-blocking
//   and will return when the tone starts, and it's up to the user to call wait()
//   or check isPlaying() between tones
// exec time(approx): 2700 cycles (~170us @ 64MHz)
//
public sub play(freq as word, duration as word = 0)
    dim prescalarbits as byte
    dim ocr as longword

    // stop any current running tone
    stop()

    // check for special case... if freq = 0 then generate a pause
    if (freq = NOTE_PAUSE) then
        delayms(duration)
        exit
    endif
    
    // compute compare value for the 16-bit timer, assuming prescaler = 1
    prescalarbits = %00             // prescaler = 1 (T1CKPS bits)
    ocr = FCYC / 2 / freq           // output compare value
    if (ocr > $ffff) then           // if we exceed range with prescaler = 1 then...
        prescalarbits = %11         // set prescaler = 8
        ocr = ocr >> 3              // and divide compare value by 8
    endif
    // compare regs are 16-bit (minus one for decr to 0 wrap)
    ocr.words(0) = ocr.words(0) - 1
    
    // set TMR prescaler
    SET_TMRx_PRESCALER(prescalarbits)

    // calculate how long to play the note in terms of the freq and duration
    // toggle_count = 2 * freq * duration / 1000
    if (duration > 0) then
        tmr_toggle_count = freq * duration / 500
    else
        tmr_toggle_count = -1
    endif

    // set the CCP output compare regs, enable CCP intr, and turn on timer
    LOAD_CCPx_COMPARE(ocr)
    TMRx_ON()

    // check to see if we're supposed to wait for the tone to finish
    if (wait_flag and (duration > 0)) then
        wait()
    endif
end sub

//
// array Play() functions
//

// const table access and misc helper functions
inline function TBLRD_() as TABLAT
    asm-
        TBLRD*
    end asm
end function

inline function TBLRD_POSTINC() as TABLAT
    asm-
        TBLRD*+
    end asm
end function

macro save_TBLPTR(dw)
    dw.bytes(2) = TBLPTRU
    dw.bytes(1) = TBLPTRH
    dw.bytes(0) = TBLPTRL
end macro

macro restore_TBLPTR(dw)
    TBLPTRU = dw.bytes(2)
    TBLPTRH = dw.bytes(1)
    TBLPTRL = dw.bytes(0)
end macro

// given a const array of (freq, duration) word note pairs, 
// compute the number of notes in the array
public macro NOTE_COUNT(array_name)
    (bound(array_name)+1)/2
end macro    

// play a const array (all or portion)
//   given a const array of (freq, duration) note pair values,
//   play 'count' number of pairs from the list
public sub play(byrefconst parray() as word, count as word)
    dim freq as word,
        duration as word,
        t_wait_flag as boolean,     // copy of wait_flag
        t_tblptr as longword        // copy of TBLPTR

    // get copy of module wait_flag
    t_wait_flag = wait_flag
    
    // set play() wait for done
    wait_flag = true

    // play each note from the list
    t_tblptr = addressof(parray)                // load TBLPTR with addr of the const string
    TBLPTRU = t_tblptr.bytes(2)
    TBLPTRH = t_tblptr.bytes(1)
    TBLPTRL = t_tblptr.bytes(0)
    while (count > 0)
        freq.bytes(0) = TBLRD_POSTINC()         // freq = parray(ix)
        freq.bytes(1) = TBLRD_POSTINC()
        duration.bytes(0) = TBLRD_POSTINC()     // duration = parray(ix+1)
        duration.bytes(1) = TBLRD_POSTINC()
        save_TBLPTR(t_tblptr)                   // save TBLPTR setting
        tone.play(freq, duration)               // play the note
        restore_TBLPTR(t_tblptr)                // restore TBLPTR setting
        count = count - 1
    end while        

    // restore wait_flag
    wait_flag = t_wait_flag
end sub

// play a const array
//   given a const array of (freq, duration) word pair values, play the notes
public sub play(byrefconst parray() as word)
    play(parray, NOTE_COUNT(parray))       // play entire list
end sub

// play a ram array (all or portion)
//   given an array of (freq, duration) word pair values in ram,
//   play 'count' number of notes from the list
public sub play(byref parray() as word, count as word)
    dim freq as word,
        duration as word,
        t_wait_flag as boolean,     // copy of wait_flag
        t_fsr as word               // copy of FSR0

    // get copy of module wait_flag
    t_wait_flag = wait_flag
    
    // set play() wait for done
    wait_flag = true

    // play each note from the list
    t_fsr = addressof(parray)               // load FSR0 with addr of parray
    FSR0 = t_fsr
    while (count > 0)
        freq.bytes(0) = POSTINC0            // freq = parray(ix)
        freq.bytes(1) = POSTINC0
        duration.bytes(0) = POSTINC0        // duration = parray(ix+1)
        duration.bytes(1) = POSTINC0
        t_fsr = FSR0                        // save FSR0
        tone.play(freq, duration)           // play note
        FSR0 = t_fsr                        // resore FSR0
        count = count - 1
    end while
    
    // restore wait_flag
    wait_flag = t_wait_flag
end sub

//
// module init
//
tmr_toggle_count = 0
wait_flag = TONE_WAIT_DEFAULT

end module
